/*
 * jMax
 * Copyright (C) 1994, 1995, 1998, 1999 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * See file LICENSE for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Based on Max/ISPW by Miller Puckette.
 *
 * Authors: Bryan Jurish <moocow@ling.uni-potsdam.de>
 *   -> much code modified from sfinfo.c, printinfo.c from libaudiofile-0.1.9 distribution
 *
 */
/* #include "fts.h" */
#include <fts/fts.h>

#include <string.h>

/* this snippet from libaudiofile-0.1.9 */
#ifdef __USE_SGI_HEADERS__
#include <dmedia/audiofile.h>
#else
#include <audiofile.h>
#endif

/* #define SFINFO_DEBUG 1 */ /* be very verbose */

typedef struct {
  fts_object_t o;   /* MUST BE FIRST STRUCTURE MEMBER */
  /* the state of the object */
# ifdef SFINFO_DEBUG
  char *filename;
# endif
  fts_atom_t fileFormat;    /* symbol: aiffc | aiff | nextsnd | wave */
  fts_atom_t nChannels;     /* int: 1 (mono) or 2 (stereo) */
  fts_atom_t frameCount;    /* float: Number of audio frames */
  fts_atom_t samplingRate;  /* float: Sampling rate in Hz */
  /* Duration in seconds = frameCount / samplingRate */
}  moo_sfinfo_t;

/* -------------------- the sfinfo object methods -------------------- */

/*-----------------------------------------------------------------
  the guts: actually read soundfile info
  -----------------------------------------------------------------*/
int moo_sfinfo_get_info (moo_sfinfo_t *inf, const char *filename)
{
  AFfilehandle  file;
  int version;  /* used by afGetFileFormat (?) */

# ifdef SFINFO_DEBUG
  inf->filename = strdup(filename);
# endif
  file = afOpenFile(filename, "r", NULL);

  /* error opening file */
  if (file == NULL) {
    return 0;
  }

  /* modified code from printinfo.c */
  switch (afGetFileFormat(file, &version)) {
  case AF_FILE_AIFFC:
    fts_set_symbol(&(inf->fileFormat), fts_new_symbol("aiffc"));
    break;
  case AF_FILE_AIFF:
    fts_set_symbol(&(inf->fileFormat), fts_new_symbol("aiff"));
    break;
  case AF_FILE_NEXTSND:
    fts_set_symbol(&(inf->fileFormat), fts_new_symbol("nextsnd"));
    break;
  case AF_FILE_WAVE:
    fts_set_symbol(&(inf->fileFormat), fts_new_symbol("wave"));
    break;
  default:
    fts_set_symbol(&(inf->fileFormat), fts_new_symbol("unknown"));
  }

  fts_set_int(&(inf->nChannels), afGetChannels(file, AF_DEFAULT_TRACK));
  fts_set_float(&(inf->frameCount), afGetFrameCount(file, AF_DEFAULT_TRACK));
  fts_set_float(&(inf->samplingRate), afGetRate(file, AF_DEFAULT_TRACK));

  afCloseFile(file);
  return 1;
}

/*-----------------------------------------------------------------
  bang on inlet: spit out data
  -----------------------------------------------------------------*/
void moo_sfinfo_bang(fts_object_t *object, int winlet, fts_symbol_t selector, 
		     int ac, const fts_atom_t *at)
{
  moo_sfinfo_t *inf = (moo_sfinfo_t *)object;

# ifdef SFINFO_DEBUG
  post("moo_sfinfo: Bang received for soundfile `%s'\n", inf->filename);
# endif

  /* output in right-to-left order as per jmax convention */
  fts_outlet_send(object, 3, fts_s_float,  1, &(inf->samplingRate));
  fts_outlet_send(object, 2, fts_s_float,  1, &(inf->frameCount));
  fts_outlet_send(object, 1, fts_s_int,    1, &(inf->nChannels));
  fts_outlet_send(object, 0, fts_s_symbol, 1, &(inf->fileFormat));

# ifdef SFINFO_DEBUG
  post("moo_sfinfo: Information sent for soundfile `%s'\n", inf->filename);
# endif

  return;
}

/*-----------------------------------------------------------------
  'set' <filename> (symbol) on inlet: just read in data
  -----------------------------------------------------------------*/
void moo_sfinfo_set(fts_object_t *object, int winlet, fts_symbol_t selector, 
		    int ac, const fts_atom_t *at)
{
  moo_sfinfo_t *inf = (moo_sfinfo_t *)object;
  fts_symbol_t fsym = fts_get_symbol_arg(ac, at, 0, fts_new_symbol(""));
  const char *filename = fts_symbol_name(fsym);

# ifdef SFINFO_DEBUG
  post("moo_sfinfo: Loading information from soundfile `%s'\n", filename);
# endif

  /* read in data */
  if (strcmp(filename,"") && !moo_sfinfo_get_info(inf, filename)) {
    post("moo_sfinfo: Error getting soundfile information for file `%s'\n", filename);
    return;
  }

# ifdef SFINFO_DEBUG
  post("moo_sfinfo: Information loaded from soundfile `%s'\n", filename);
# endif
  return;
}

/*-----------------------------------------------------------------
  'load' <filename> (symbol) on inlet: read in & spit out data
  -----------------------------------------------------------------*/
void moo_sfinfo_load(fts_object_t *object, int winlet, fts_symbol_t selector, 
		     int ac, const fts_atom_t *at)
{
  moo_sfinfo_set(object, winlet, selector, ac, at);
  moo_sfinfo_bang(object, winlet, selector, ac, at);
  return;
}



/*-----------------------------------------------------------------
  instance initialization method
  -----------------------------------------------------------------*/
void moo_sfinfo_init(fts_object_t *object, int winlet, fts_symbol_t selector, 
		     int ac, const fts_atom_t *at)
{
  moo_sfinfo_t *inf = (moo_sfinfo_t *)object;
  fts_symbol_t fsym = fts_get_symbol_arg(ac, at, 1, fts_new_symbol(""));
  const char *filename = fts_symbol_name(fsym);

  /* read in data */
  if (strcmp(filename,"") && !moo_sfinfo_get_info(inf, filename)) {
    post("moo_sfinfo: Error getting soundfile information for file `%s'\n", filename);
  }
  return;
}


/*-----------------------------------------------------------------
  class instantiation function
  -----------------------------------------------------------------*/
static fts_status_t moo_sfinfo_instantiate(fts_class_t *class, int ac, const fts_atom_t *at)
{
  fts_type_t s,i,f,ary[2];

# ifdef SFINFO_DEBUG
  post( "Instantiating class `moo_sfinfo'\n");
# endif
  
  s = fts_t_symbol;
  i = fts_t_int;
  f = fts_t_float;

  ary[0] = s;
  ary[1] = s;

  /* intialize the class */
  fts_class_init(class, sizeof(moo_sfinfo_t), 1, 4, NULL);

  /* method definitions */
  fts_method_define_optargs(class, fts_SystemInlet, fts_s_init,
			    moo_sfinfo_init, 2, ary, 1);
  fts_method_define(class, 0, fts_s_bang, moo_sfinfo_bang, 0, NULL);
  fts_method_define(class, 0, fts_s_set,  moo_sfinfo_set,  1, &s);
  fts_method_define(class, 0, fts_s_load, moo_sfinfo_load, 1, &s);
  
  /* static outlet typing */
  fts_outlet_type_define(class, 0, fts_s_symbol, 1, &s); /* fileFormat */
  fts_outlet_type_define(class, 1, fts_s_int,    1, &i); /* nChannels  */
  fts_outlet_type_define(class, 2, fts_s_float,  1, &f); /* frameCount */
  fts_outlet_type_define(class, 3, fts_s_float,  1, &f); /* sampleRate */

  return fts_Success;
}


/*-----------------------------------------------------------------
  class installation function
  -----------------------------------------------------------------*/
void moo_sfinfo_config(void)
{
# ifdef SFINFO_DEBUG
  post( "Installing class `moo_sfinfo'\n");
# endif

  /* Use a bizarre class name no to overwrite any (possible future) standard class(es) */
  fts_class_install( fts_new_symbol("moo_sfinfo"), moo_sfinfo_instantiate);

  /* ... and register an alias: "sfinfo" */
  fts_class_alias( fts_new_symbol("sfinfo"), fts_new_symbol("moo_sfinfo"));
}

/*-----------------------------------------------------------------
  black magic (from Christian Klippel's filters.c)
  -----------------------------------------------------------------*/
fts_module_t sfinfo_module = { "sfinfo", "audio file information objects", moo_sfinfo_config, 0, 0};
